import importlib.resources

import pytest

from grafana_sync.api.client import FOLDER_GENERAL, GrafanaClient
from grafana_sync.api.models import (
    DashboardData,
    DatasourceDefinition,
    GetDashboardResponse,
)
from grafana_sync.dashboards.models import DataSource
from grafana_sync.exceptions import DestinationParentNotFoundError, GrafanaApiError
from grafana_sync.sync import sync

from . import dashboards, responses

pytestmark = pytest.mark.docker


def read_db(filename: str) -> DashboardData:
    ref = importlib.resources.files(dashboards) / filename
    with importlib.resources.as_file(ref) as path, open(path, "rb") as f:
        return DashboardData.model_validate_json(f.read())


def read_response(filename: str) -> GetDashboardResponse:
    ref = importlib.resources.files(responses) / filename
    with importlib.resources.as_file(ref) as path, open(path, "rb") as f:
        return GetDashboardResponse.model_validate_json(f.read())


async def test_sync_dashboard(grafana: GrafanaClient, grafana_dst: GrafanaClient):
    dashboard1 = DashboardData(uid="dash1", title="Dashboard 1")

    await grafana.update_dashboard(dashboard1)

    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
    )

    dst_db = await grafana_dst.get_dashboard("dash1")
    assert dst_db.dashboard.title == "Dashboard 1"


async def test_sync_dashboard_with_ds_migration(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    dashboard1 = read_db("simple-ds-var.json")

    await grafana.update_dashboard(dashboard1)

    await grafana.create_datasource(
        DatasourceDefinition(
            name="prometheus",
            uid="P1809F7CD0C75ACF3",
            type="prometheus",
            access="proxy",
        )
    )

    await grafana_dst.create_datasource(
        DatasourceDefinition(
            name="prometheus-dst",
            uid="my-new-uid",
            type="prometheus",
            access="proxy",
        )
    )

    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        migrate_datasources=True,
    )

    dst_db = await grafana_dst.get_dashboard("simple-ds-var")
    assert dst_db.dashboard.title == "simple-ds-var"

    assert dst_db.dashboard.templating is not None
    assert dst_db.dashboard.templating.list_ is not None
    assert dst_db.dashboard.templating.list_[0].current is not None

    assert dst_db.dashboard.templating.list_[0].current.text == "prometheus-dst"
    assert dst_db.dashboard.templating.list_[0].current.value == "my-new-uid"


async def test_sync_dashboard_with_classic_ds_migration_datasource_string(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    dashboard1 = read_response("get-dashboard-datasource-string.json").dashboard

    await grafana.update_dashboard(dashboard1)

    await grafana.create_datasource(
        DatasourceDefinition(
            name="InfluxDB Produktion Telegraf",
            uid="old-uid",
            type="influxdb",
            access="proxy",
        )
    )

    await grafana_dst.create_datasource(
        DatasourceDefinition(
            name="influxdb-dst",
            uid="my-new-uid",
            type="influxdb",
            access="proxy",
        )
    )

    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        migrate_datasources=True,
    )

    dst_db = await grafana_dst.get_dashboard("datasource-string")
    assert dst_db.dashboard.title == "datasource-string"

    assert dst_db.dashboard.panels is not None
    assert dst_db.dashboard.panels[0].datasource is not None
    assert isinstance(dst_db.dashboard.panels[0].datasource, DataSource)
    assert dst_db.dashboard.panels[0].datasource.type_ == "influxdb"
    assert dst_db.dashboard.panels[0].datasource.uid == "my-new-uid"


async def test_sync_dashboard_with_classic_ds_migration_panel_target(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    dashboard1 = read_response("get-dashboard-panel-target.json").dashboard

    await grafana.update_dashboard(dashboard1)

    await grafana.create_datasource(
        DatasourceDefinition(
            name="My InfluxDB DataSource",
            uid="influx",
            type="influxdb",
            access="proxy",
        )
    )

    await grafana_dst.create_datasource(
        DatasourceDefinition(
            name="influxdb-dst",
            uid="my-new-uid",
            type="influxdb",
            access="proxy",
        )
    )

    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        migrate_datasources=True,
    )

    dst_db = await grafana_dst.get_dashboard("panel-target")
    assert dst_db.dashboard.title == "panel-target"

    assert dst_db.dashboard.panels is not None
    assert dst_db.dashboard.panels[0].datasource is not None
    assert isinstance(dst_db.dashboard.panels[0].datasource, DataSource)
    assert dst_db.dashboard.panels[0].datasource.type_ == "influxdb"
    assert dst_db.dashboard.panels[0].datasource.uid == "my-new-uid"


async def test_sync_folder(grafana: GrafanaClient, grafana_dst: GrafanaClient):
    await grafana.create_folder(title="Folder 1", uid="folder1")
    dashboard1 = DashboardData(uid="dash1", title="Dashboard 1")

    await grafana.update_dashboard(dashboard1, folder_uid="folder1")

    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
    )

    dst_db = await grafana_dst.get_dashboard("dash1")
    assert dst_db.dashboard.title == "Dashboard 1"
    assert dst_db.meta.folder_uid == "folder1"

    dst_folder = await grafana_dst.get_folder("folder1")
    assert dst_folder.parent_uid is None
    assert dst_folder.title == "Folder 1"


async def test_sync_folder_relocation(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    # Create parent folders
    await grafana.create_folder(title="Parent 1", uid="parent1")
    await grafana.create_folder(title="Parent 2", uid="parent2")

    # Create child folder in Parent 1
    await grafana.create_folder(title="Child", uid="child", parent_uid="parent1")

    # Create same structure in destination, but with Child under Parent 1
    await grafana_dst.create_folder(title="Parent 1", uid="parent1")
    await grafana_dst.create_folder(title="Parent 2", uid="parent2")
    await grafana_dst.create_folder(title="Child", uid="child", parent_uid="parent1")

    # Move child folder to Parent 2 in source
    await grafana.move_folder("child", "parent2")

    # Verify folder was moved
    src_child = await grafana.get_folder("child")
    assert src_child.parent_uid == "parent2"

    # Sync should move the folder in destination
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        recursive=True,
    )

    # Verify folder was moved
    dst_child = await grafana_dst.get_folder("child")
    assert dst_child.parent_uid == "parent2"


async def test_sync_folder_no_relocation(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    # Create parent folders
    await grafana.create_folder(title="Parent 1", uid="parent1")
    await grafana.create_folder(title="Parent 2", uid="parent2")

    # Create child folder in Parent 1
    await grafana.create_folder(title="Child", uid="child", parent_uid="parent1")

    # Create same structure in destination, but with Child under Parent 1
    await grafana_dst.create_folder(title="Parent 1", uid="parent1")
    await grafana_dst.create_folder(title="Parent 2", uid="parent2")
    await grafana_dst.create_folder(title="Child", uid="child", parent_uid="parent1")

    # Move child folder to Parent 2 in source
    await grafana.move_folder("child", "parent2")

    # Verify folder was moved in source
    src_child = await grafana.get_folder("child")
    assert src_child.parent_uid == "parent2"

    # Sync with relocate_folders=False should not move the folder in destination
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        recursive=True,
        relocate_folders=False,
    )

    # Verify folder was NOT moved in destination
    dst_child = await grafana_dst.get_folder("child")
    assert dst_child.parent_uid == "parent1"


async def test_sync_dashboard_relocation(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    # Create folders in source and destination
    await grafana.create_folder(title="Folder 1", uid="folder1")
    await grafana.create_folder(title="Folder 2", uid="folder2")
    await grafana_dst.create_folder(title="Folder 1", uid="folder1")
    await grafana_dst.create_folder(title="Folder 2", uid="folder2")

    # Create dashboard in Folder 1
    dashboard = DashboardData(uid="dash1", title="Dashboard 1")
    await grafana.update_dashboard(dashboard, folder_uid="folder1")
    await grafana_dst.update_dashboard(dashboard, folder_uid="folder1")

    # Move dashboard to Folder 2 in source
    dashboard = DashboardData(uid="dash1", title="Dashboard 1")
    await grafana.update_dashboard(dashboard, folder_uid="folder2")

    # Verify dashboard was moved
    src_db = await grafana.get_dashboard("dash1")
    assert src_db.meta.folder_uid == "folder2"

    # Sync should move the dashboard in destination
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
    )

    # Verify dashboard was moved
    dst_db = await grafana_dst.get_dashboard("dash1")
    assert dst_db.meta.folder_uid == "folder2"


async def test_sync_selected_folder(grafana: GrafanaClient, grafana_dst: GrafanaClient):
    await grafana.create_folder(title="Folder 1", uid="folder1")
    await grafana.create_folder(title="Folder 2", uid="folder2")
    dashboard1 = DashboardData(uid="dash1", title="Dashboard 1")
    dashboard2 = DashboardData(uid="dash2", title="Dashboard 2")
    dashboard3 = DashboardData(uid="dash3", title="Dashboard 3")

    await grafana.update_dashboard(dashboard1, folder_uid="folder1")
    await grafana.update_dashboard(dashboard2, folder_uid="folder2")
    await grafana.update_dashboard(dashboard3)  # general

    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        folder_uid="folder1",
    )

    dst_db = await grafana_dst.get_dashboard("dash1")
    assert dst_db.dashboard.title == "Dashboard 1"
    assert dst_db.meta.folder_uid == "folder1"

    # ensure nothing else was synced
    await grafana_dst.delete_folder("folder1")
    await grafana_dst.check_pristine()


async def test_sync_dashboard_no_relocation(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    # Create folders in source and destination
    await grafana.create_folder(title="Folder 1", uid="folder1")
    await grafana.create_folder(title="Folder 2", uid="folder2")
    await grafana_dst.create_folder(title="Folder 1", uid="folder1")
    await grafana_dst.create_folder(title="Folder 2", uid="folder2")

    # Create dashboard in Folder 1
    dashboard = DashboardData(uid="dash1", title="Dashboard 1")
    await grafana.update_dashboard(dashboard, folder_uid="folder1")
    await grafana_dst.update_dashboard(dashboard, folder_uid="folder1")

    # Move dashboard to Folder 2 in source
    dashboard = DashboardData(uid="dash1", title="Dashboard 1")
    await grafana.update_dashboard(dashboard, folder_uid="folder2")

    # Verify dashboard was moved in source
    src_db = await grafana.get_dashboard("dash1")
    assert src_db.meta.folder_uid == "folder2"

    # Sync with relocate_dashboards=False should not move the dashboard in destination
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        relocate_dashboards=False,
    )

    # Get version before sync
    dst_db_before = await grafana_dst.get_dashboard("dash1")
    version_before = dst_db_before.dashboard.version

    # Verify dashboard was NOT moved in destination and version didn't change
    dst_db_after = await grafana_dst.get_dashboard("dash1")
    assert dst_db_after.meta.folder_uid == "folder1"
    assert (
        dst_db_after.dashboard.version == version_before
    )  # Version should not increase


async def test_sync_to_destination_parent(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    # Create source structure
    await grafana.create_folder(title="Folder 1", uid="folder1")
    await grafana.create_folder(title="Folder 2", uid="folder2")
    await grafana.create_folder(title="Child", uid="child", parent_uid="folder1")

    # Create destination parent folder
    await grafana_dst.create_folder(title="Destination Parent", uid="dst_parent")

    # Create some dashboards
    dashboard1 = DashboardData(uid="dash1", title="Dashboard 1")
    dashboard2 = DashboardData(uid="dash2", title="Dashboard 2")
    dashboard3 = DashboardData(uid="dash3", title="Dashboard 3")

    await grafana.update_dashboard(dashboard1, folder_uid="folder1")
    await grafana.update_dashboard(dashboard2, folder_uid="folder2")
    await grafana.update_dashboard(dashboard3, folder_uid="child")

    # Sync everything under the destination parent
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        dst_parent_uid="dst_parent",
    )

    # Verify folders were created under destination parent
    dst_folder1 = await grafana_dst.get_folder("folder1")
    assert dst_folder1.parent_uid == "dst_parent"

    dst_folder2 = await grafana_dst.get_folder("folder2")
    assert dst_folder2.parent_uid == "dst_parent"

    dst_child = await grafana_dst.get_folder("child")
    assert (
        dst_child.parent_uid == "folder1"
    )  # Should maintain hierarchy under new parent

    # Verify dashboards are in correct folders
    dash1 = await grafana_dst.get_dashboard("dash1")
    assert dash1.meta.folder_uid == "folder1"

    dash2 = await grafana_dst.get_dashboard("dash2")
    assert dash2.meta.folder_uid == "folder2"

    dash3 = await grafana_dst.get_dashboard("dash3")
    assert dash3.meta.folder_uid == "child"


async def test_sync_to_general_folder(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    # Create source structure with nested folders
    await grafana.create_folder(title="Parent", uid="parent")
    await grafana.create_folder(title="Child", uid="child", parent_uid="parent")

    # Create a dashboard in the child folder
    dashboard = DashboardData(uid="dash1", title="Dashboard 1")
    await grafana.update_dashboard(dashboard, folder_uid="child")

    # Sync with dst_parent_uid set to FOLDER_GENERAL
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        dst_parent_uid=FOLDER_GENERAL,
    )

    # Verify parent folder is at root level
    dst_parent = await grafana_dst.get_folder("parent")
    assert dst_parent.parent_uid is None

    # Verify child folder maintains its hierarchy
    dst_child = await grafana_dst.get_folder("child")
    assert dst_child.parent_uid == "parent"

    # Verify dashboard is in correct folder
    dash = await grafana_dst.get_dashboard("dash1")
    assert dash.meta.folder_uid == "child"


async def test_sync_to_nonexistent_parent(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    # Create source structure with nested folders
    await grafana.create_folder(title="Parent", uid="parent")
    await grafana.create_folder(title="Child", uid="child", parent_uid="parent")

    # Create a dashboard in the child folder
    dashboard = DashboardData(uid="dash1", title="Dashboard 1")
    await grafana.update_dashboard(dashboard, folder_uid="child")

    # Attempt sync with non-existent destination parent
    with pytest.raises(DestinationParentNotFoundError) as exc_info:
        await sync(
            src_grafana=grafana,
            dst_grafana=grafana_dst,
            dst_parent_uid="nonexistent",
        )

    assert exc_info.value.parent_uid == "nonexistent"
    assert (
        str(exc_info.value)
        == "Destination parent folder with UID 'nonexistent' does not exist"
    )

    # Verify nothing was synced
    await grafana_dst.check_pristine()


async def test_sync_root_dashboard_to_destination_parent(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    """Test that a dashboard at root level is moved to dst_parent_uid folder when specified."""
    # Create a dashboard at root level in source
    dashboard = DashboardData(uid="dash1", title="Root Dashboard")
    await grafana.update_dashboard(dashboard)  # No folder_uid = root level

    # Create destination parent folder
    await grafana_dst.create_folder(title="Destination Parent", uid="dst_parent")

    # Sync with dst_parent_uid
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        dst_parent_uid="dst_parent",
    )

    # Verify dashboard was moved to destination parent folder
    dst_dash = await grafana_dst.get_dashboard("dash1")
    assert dst_dash.meta.folder_uid == "dst_parent"
    assert dst_dash.dashboard.title == "Root Dashboard"


async def test_sync_with_pruning_and_destination_parent(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    """Test that pruning works correctly when a destination parent is specified."""
    # Create source structure
    await grafana.create_folder(title="Source Folder", uid="src_folder")
    dashboard1 = DashboardData(uid="dash1", title="Dashboard 1")
    dashboard2 = DashboardData(uid="dash2", title="Dashboard 2")
    await grafana.update_dashboard(dashboard1, folder_uid="src_folder")
    await grafana.update_dashboard(dashboard2, folder_uid="src_folder")

    # Create destination parent and structure
    await grafana_dst.create_folder(title="Destination Parent", uid="dst_parent")
    await grafana_dst.create_folder(
        title="Source Folder", uid="src_folder", parent_uid="dst_parent"
    )

    # Create extra dashboard in destination that should be pruned
    dashboard3 = DashboardData(uid="dash3", title="Dashboard 3")
    await grafana_dst.update_dashboard(dashboard3, folder_uid="src_folder")

    # Sync with pruning enabled and destination parent specified
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        folder_uid="src_folder",
        prune=True,
        dst_parent_uid="dst_parent",
    )

    # Verify dashboards 1 and 2 exist in destination under the correct folder
    dst_db1 = await grafana_dst.get_dashboard("dash1")
    assert dst_db1.dashboard.title == "Dashboard 1"
    assert dst_db1.meta.folder_uid == "src_folder"

    dst_db2 = await grafana_dst.get_dashboard("dash2")
    assert dst_db2.dashboard.title == "Dashboard 2"
    assert dst_db2.meta.folder_uid == "src_folder"

    # Verify dashboard 3 was pruned
    with pytest.raises(GrafanaApiError):
        await grafana_dst.get_dashboard("dash3")

    # Verify folder structure
    dst_folder = await grafana_dst.get_folder("src_folder")
    assert dst_folder.parent_uid == "dst_parent"


async def test_move_folders_with_destination_parent(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    """Test that folders are correctly moved when a destination parent is specified."""
    # Create source structure with nested folders
    await grafana.create_folder(title="Parent", uid="parent")
    await grafana.create_folder(title="Child 1", uid="child1", parent_uid="parent")
    await grafana.create_folder(title="Child 2", uid="child2")  # At root level
    await grafana.create_folder(
        title="Grandchild", uid="grandchild", parent_uid="child1"
    )

    # Create destination parent and initial structure (with different hierarchy)
    await grafana_dst.create_folder(title="Destination Parent", uid="dst_parent")
    await grafana_dst.create_folder(
        title="Parent", uid="parent", parent_uid="dst_parent"
    )
    await grafana_dst.create_folder(title="Child 1", uid="child1")  # At root initially
    await grafana_dst.create_folder(title="Child 2", uid="child2")  # At root
    await grafana_dst.create_folder(
        title="Grandchild", uid="grandchild", parent_uid="child2"
    )  # Wrong parent

    # Sync with destination parent
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        dst_parent_uid="dst_parent",
    )

    # Verify final folder structure
    dst_parent = await grafana_dst.get_folder("parent")
    assert dst_parent.parent_uid == "dst_parent"

    dst_child1 = await grafana_dst.get_folder("child1")
    assert dst_child1.parent_uid == "parent"

    dst_child2 = await grafana_dst.get_folder("child2")
    assert dst_child2.parent_uid == "dst_parent"  # Should be under dst_parent

    dst_grandchild = await grafana_dst.get_folder("grandchild")
    assert dst_grandchild.parent_uid == "child1"  # Should match source hierarchy


async def test_sync_existing_general_dashboard_to_destination_parent(
    grafana: GrafanaClient, grafana_dst: GrafanaClient
):
    """Test that an existing dashboard in general folder is moved when dst_parent_uid is set."""
    # Create a dashboard at root level in both source and destination
    dashboard = DashboardData(uid="dash1", title="Root Dashboard")
    await grafana.update_dashboard(dashboard)  # No folder_uid = root level
    await grafana_dst.update_dashboard(dashboard)  # Also at root level

    # Verify initial state - dashboard exists at root in both instances
    src_dash = await grafana.get_dashboard("dash1")
    assert src_dash.meta.folder_uid == ""
    dst_dash = await grafana_dst.get_dashboard("dash1")
    assert dst_dash.meta.folder_uid == ""

    # Create destination parent folder
    await grafana_dst.create_folder(title="Destination Parent", uid="dst_parent")

    # Sync with dst_parent_uid
    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        dst_parent_uid="dst_parent",
    )

    # Verify dashboard was moved to destination parent folder
    dst_dash_after = await grafana_dst.get_dashboard("dash1")
    assert dst_dash_after.meta.folder_uid == "dst_parent"
    assert dst_dash_after.dashboard.title == "Root Dashboard"

    # Verify source dashboard remains at root
    src_dash_after = await grafana.get_dashboard("dash1")
    assert src_dash_after.meta.folder_uid == ""


async def test_sync_with_pruning(grafana: GrafanaClient, grafana_dst: GrafanaClient):
    # Create folders in source and destination
    await grafana.create_folder(title="Folder 1", uid="folder1")
    await grafana_dst.create_folder(title="Folder 1", uid="folder1")

    # Create dashboards in source
    dashboard1 = DashboardData(uid="dash1", title="Dashboard 1")
    dashboard2 = DashboardData(uid="dash2", title="Dashboard 2")
    await grafana.update_dashboard(dashboard1, folder_uid="folder1")
    await grafana.update_dashboard(dashboard2, folder_uid="folder1")

    # Create extra dashboard in destination that should be pruned
    dashboard3 = DashboardData(uid="dash3", title="Dashboard 3")
    await grafana_dst.update_dashboard(dashboard3, folder_uid="folder1")

    await sync(
        src_grafana=grafana,
        dst_grafana=grafana_dst,
        folder_uid="folder1",
        prune=True,
    )

    # Verify dashboards 1 and 2 exist in destination
    dst_db1 = await grafana_dst.get_dashboard("dash1")
    assert dst_db1.dashboard.title == "Dashboard 1"
    dst_db2 = await grafana_dst.get_dashboard("dash2")
    assert dst_db2.dashboard.title == "Dashboard 2"

    # Verify dashboard 3 was pruned
    try:
        await grafana_dst.get_dashboard("dash3")
        raise AssertionError("Dashboard 3 should have been pruned")
    except Exception:
        pass



import importlib.resources

import pytest

from grafana_sync.api.models import GetDashboardResponse
from grafana_sync.dashboards.models import (
    DashboardData,
    DataSource,
    DSRef,
    Panel,
    Target,
)

from . import dashboards, responses


def read_db(filename: str) -> DashboardData:
    ref = importlib.resources.files(dashboards) / filename
    with importlib.resources.as_file(ref) as path, open(path, "rb") as f:
        return DashboardData.model_validate_json(f.read())


def read_response(filename: str) -> GetDashboardResponse:
    ref = importlib.resources.files(responses) / filename
    with importlib.resources.as_file(ref) as path, open(path, "rb") as f:
        return GetDashboardResponse.model_validate_json(f.read())


def test_ds_inheritance():
    db = DashboardData(
        uid="test",
        title="test",
        panels=[
            Panel(
                datasource=DataSource(
                    type="influxdb",
                    uid="orig-uid",
                ),
                targets=[
                    Target(
                        datasource=DataSource(uid="${DataSource}"),
                    )
                ],
            )
        ],
    )

    db.update_datasources({"orig-uid": DSRef(uid="new-uid", name="InfluxDB")})

    assert db == DashboardData(
        uid="test",
        title="test",
        panels=[
            Panel(
                datasource=DataSource(
                    type="influxdb",
                    uid="new-uid",
                ),
                targets=[
                    Target(
                        datasource=DataSource(uid="${DataSource}"),
                    )
                ],
            )
        ],
    )


@pytest.mark.parametrize(
    ("filename", "ct"),
    [
        ("haproxy-2-full.json", 0),
        ("host-overview.json", 0),
        ("simple-ds-var.json", 1),
        ("simple-novar.json", 2),
    ],
)
def test_update_datasources(filename, ct):
    db = read_db(filename)

    assert ct == db.update_datasources(
        {"P1809F7CD0C75ACF3": DSRef(uid="foobar", name="foobar")}
    )


@pytest.mark.parametrize(
    ("filename", "ct"),
    [
        ("get-dashboard-datasource-string.json", 1),
        ("get-dashboard-panel-target.json", 1),
    ],
)
def test_update_classic_datasource_from_response(filename, ct):
    res = read_response(filename)
    db = res.dashboard

    ds_config = {
        "InfluxDB Produktion Telegraf": DataSource(
            type="influxdb", uid="influxdb-prod-telegraf"
        )
    }

    db.upgrade_datasources(ds_config)

    assert ct == db.update_datasources(
        {
            "influxdb-prod-telegraf": DSRef(uid="foobar", name="foobar"),
            "influx": DSRef(uid="foobar", name="foobar"),
        }
    )



import logging
from collections.abc import Mapping
from typing import TYPE_CHECKING

from grafana_sync.api.client import FOLDER_GENERAL, FOLDER_SHAREDWITHME
from grafana_sync.api.models import DatasourceDefinition
from grafana_sync.dashboards.models import DataSource, DSRef
from grafana_sync.datasource_mapper import map_datasources
from grafana_sync.exceptions import DestinationParentNotFoundError

if TYPE_CHECKING:
    from grafana_sync.api.client import GrafanaClient
    from grafana_sync.api.models import DashboardData

logger = logging.getLogger(__name__)


class GrafanaSync:
    """Handles synchronization of folders and dashboards between Grafana instances."""

    ds_map: Mapping[str, DSRef] | None
    src_datasources: list[DatasourceDefinition] | None
    dst_datasources: list[DatasourceDefinition] | None
    src_ds_config: Mapping[str, DataSource] | None

    def __init__(
        self,
        src_grafana: "GrafanaClient",
        dst_grafana: "GrafanaClient",
        *,
        dst_parent_uid: str | None = None,
        migrate_datasources: bool = False,
    ) -> None:
        self.src_grafana = src_grafana
        self.dst_grafana = dst_grafana
        self.folder_relocation_queue: dict[str, str] = {}
        self.dst_parent_uid = dst_parent_uid
        self.migrate_datasources = migrate_datasources
        self.ds_map = None
        self.src_datasources = None
        self.dst_datasources = None
        self.src_ds_config = None

    async def sync_folder(
        self,
        folder_uid: str,
        can_move: bool,
    ) -> None:
        """Sync a single folder from source to destination Grafana instance."""
        src_folder = await self.src_grafana.get_folder(folder_uid)
        title = src_folder.title
        parent_uid = src_folder.parent_uid or self.dst_parent_uid or FOLDER_GENERAL

        # Check if folder already exists
        try:
            existing_dst_folder = await self.dst_grafana.get_folder(folder_uid)
        except Exception:
            # Folder doesn't exist, create it
            logger.info("Creating folder '%s' in destination", title)
            try:
                # Handle dst_parent_uid for top-level folders
                if parent_uid == FOLDER_GENERAL:
                    if self.dst_parent_uid == FOLDER_GENERAL:
                        dst_parent_uid = None  # Explicitly place at root
                    elif self.dst_parent_uid is not None:
                        dst_parent_uid = self.dst_parent_uid
                    else:
                        dst_parent_uid = None
                else:
                    # Check if parent_uid is available in dst
                    try:
                        await self.dst_grafana.get_folder(parent_uid)
                    except Exception:
                        dst_parent_uid = None
                    else:
                        dst_parent_uid = parent_uid

                await self.dst_grafana.create_folder(
                    title=title, uid=folder_uid, parent_uid=dst_parent_uid
                )
                logger.info("Created folder '%s' (uid: %s)", title, folder_uid)
            except Exception:
                logger.exception("Failed to create folder '%s'", title)
        else:
            if existing_dst_folder.title != title:
                logger.info("Updating folder title '%s' in destination", title)
                try:
                    await self.dst_grafana.update_folder(
                        uid=folder_uid,
                        title=title,
                        overwrite=True,
                    )
                except Exception:
                    logger.exception("Failed to update folder '%s'", title)

            # check if the folder needs to be moved
            if (
                can_move
                and (existing_dst_folder.parent_uid or FOLDER_GENERAL) != parent_uid
            ):
                # since a parent might not exist yet, we enqueue the relocations
                self.folder_relocation_queue[folder_uid] = parent_uid

    async def move_folders_to_new_parents(self) -> None:
        for folder_uid, parent_uid in self.folder_relocation_queue.items():
            if folder_uid in [FOLDER_GENERAL, FOLDER_SHAREDWITHME]:
                continue  # skip system folders

            try:
                await self.dst_grafana.move_folder(
                    folder_uid, parent_uid if parent_uid != FOLDER_GENERAL else None
                )
            except Exception:
                logger.exception(
                    "Failed to move folder '%s' to new parent '%s'",
                    folder_uid,
                    parent_uid,
                )
            else:
                logger.info(
                    "Moved folder '%s' to new parent '%s'", folder_uid, parent_uid
                )

        self.folder_relocation_queue.clear()

    async def get_folder_dashboards(
        self,
        grafana: "GrafanaClient",
        folder_uid: str,
        recursive: bool,
    ) -> set[str]:
        """Get all dashboard UIDs in a folder (and optionally its subfolders)."""
        dashboard_uids = set()

        try:
            async for _, _, dashboards in grafana.walk(
                folder_uid, recursive, include_dashboards=True
            ):
                for dashboard in dashboards.root:
                    dashboard_uids.add(dashboard.uid)
        except Exception as e:
            logger.warning("Failed to get dashboards for folder %s: %s", folder_uid, e)

        return dashboard_uids

    async def delete_dashboard(self, dashboard_uid: str) -> bool:
        """Delete a dashboard from destination Grafana instance."""
        try:
            await self.dst_grafana.delete_dashboard(dashboard_uid)
        except Exception:
            logger.exception("Failed to delete dashboard %s", dashboard_uid)
            return False
        else:
            logger.info("Deleted dashboard with uid: %s", dashboard_uid)
            return True

    def _clean_dashboard_for_comparison(self, dashboard_data: "DashboardData") -> dict:
        """Remove dynamic fields from dashboard data for comparison."""
        return dashboard_data.model_dump(exclude={"id", "version"}, by_alias=True)

    async def get_src_datasources(self):
        if self.src_datasources is not None:
            return self.src_datasources

        self.src_datasources = (await self.src_grafana.get_datasources()).root
        return self.src_datasources

    async def get_dst_datasources(self):
        if self.dst_datasources is not None:
            return self.dst_datasources

        self.dst_datasources = (await self.dst_grafana.get_datasources()).root
        return self.dst_datasources

    async def get_ds_map(self) -> Mapping[str, DSRef]:
        if self.ds_map is not None:
            return self.ds_map

        self.ds_map = map_datasources(
            await self.get_src_datasources(),
            await self.get_dst_datasources(),
        )

        logger.info("Mapped datasources: %s", self.ds_map)

        return self.ds_map

    async def get_src_ds_config(self) -> Mapping[str, DataSource]:
        if self.src_ds_config is not None:
            return self.src_ds_config

        self.src_ds_config = {
            ds.name: DataSource(
                type=ds.type_,
                uid=ds.uid,
            )
            for ds in (await self.get_src_datasources())
        }

        return self.src_ds_config

    async def sync_dashboard(
        self,
        dashboard_uid: str,
        folder_uid: str | None = None,
        relocate=True,
    ) -> bool:
        """Sync a single dashboard from source to destination Grafana instance.

        Returns True if sync was successful, False otherwise.
        """
        try:
            # Get dashboard from source
            src_dashboard = await self.src_grafana.get_dashboard(dashboard_uid)
            if not src_dashboard:
                logger.error("Dashboard %s not found in source", dashboard_uid)
                return False

            src_data = src_dashboard.dashboard

            if self.migrate_datasources:
                ds_map = await self.get_ds_map()
                src_ds_config = await self.get_src_ds_config()
                src_data.upgrade_datasources(src_ds_config)
                src_data.update_datasources(ds_map)

            if folder_uid == FOLDER_GENERAL:
                target_folder = (
                    self.dst_parent_uid
                    if self.dst_parent_uid != FOLDER_GENERAL
                    else None
                )
            else:
                target_folder = folder_uid

            # Check if dashboard exists in destination
            try:
                dst_dashboard = await self.dst_grafana.get_dashboard(dashboard_uid)
            except Exception:
                # Dashboard doesn't exist in destination
                dst_dashboard = None
            else:
                # Compare dashboards after cleaning
                if self._clean_dashboard_for_comparison(
                    src_data
                ) == self._clean_dashboard_for_comparison(dst_dashboard.dashboard) and (
                    target_folder == dst_dashboard.meta.folder_uid or not relocate
                ):
                    logger.info(
                        "Dashboard '%s' (uid: %s) is identical, skipping update",
                        src_data.title,
                        dashboard_uid,
                    )
                    return True

            if dst_dashboard is not None and not relocate:
                target_folder = dst_dashboard.meta.folder_uid

            await self.dst_grafana.update_dashboard(
                src_data,
                folder_uid=target_folder,
            )
        except Exception:
            logger.exception("Failed to sync dashboard %s", dashboard_uid)
            return False
        else:
            logger.info(
                "Synced dashboard '%s' (uid: %s)",
                src_data.title,
                dashboard_uid,
            )
            return True


async def sync(
    *,
    src_grafana: "GrafanaClient",
    dst_grafana: "GrafanaClient",
    folder_uid: str = FOLDER_GENERAL,
    recursive: bool = True,
    include_dashboards: bool = True,
    prune: bool = False,
    relocate_folders: bool = True,
    relocate_dashboards: bool = True,
    dst_parent_uid: str | None = None,
    migrate_datasources: bool = False,
):
    # Verify destination parent exists if specified
    if dst_parent_uid is not None and dst_parent_uid != FOLDER_GENERAL:
        try:
            await dst_grafana.get_folder(dst_parent_uid)
        except Exception as e:
            raise DestinationParentNotFoundError(dst_parent_uid) from e

    syncer = GrafanaSync(
        src_grafana,
        dst_grafana,
        dst_parent_uid=dst_parent_uid,
        migrate_datasources=migrate_datasources,
    )

    # Track source dashboards if pruning is enabled
    src_dashboard_uids = set()
    dst_dashboard_uids = set()

    if include_dashboards and prune:
        # Get all dashboards in destination folders before we start syncing
        dst_dashboard_uids = await syncer.get_folder_dashboards(
            dst_grafana, folder_uid, recursive
        )

    # if a folder was requested sync it first
    if folder_uid != FOLDER_GENERAL:
        await syncer.sync_folder(folder_uid, can_move=False)

    # Now walk and sync child folders and optionally dashboards
    async for root_uid, folders, dashboards in src_grafana.walk(
        folder_uid, recursive, include_dashboards=include_dashboards
    ):
        for folder in folders.root:
            if folder == FOLDER_SHAREDWITHME:
                continue  # skip unsyncable folder

            await syncer.sync_folder(folder.uid, can_move=True)

        # Sync dashboards if requested
        if include_dashboards:
            for dashboard in dashboards.root:
                dashboard_uid = dashboard.uid
                if await syncer.sync_dashboard(
                    dashboard_uid, root_uid, relocate=relocate_dashboards
                ):
                    src_dashboard_uids.add(dashboard_uid)

    if relocate_folders:
        logger.info("relocation folders to updated parents if needed")
        await syncer.move_folders_to_new_parents()
    else:
        logger.info("skipping folder relocation (disabled)")

    # Prune dashboards that don't exist in source
    if include_dashboards and prune:
        dashboards_to_delete = dst_dashboard_uids - src_dashboard_uids
        for dashboard_uid in dashboards_to_delete:
            await syncer.delete_dashboard(dashboard_uid)



from collections.abc import Generator, Mapping

from pydantic import BaseModel, ConfigDict, Field

from grafana_sync.exceptions import UnmappedDatasourceError


class DSRef(BaseModel):
    uid: str
    name: str


class DataSource(BaseModel):
    type_: str | None = Field(alias="type", default=None)
    uid: str

    model_config = ConfigDict(extra="allow")

    @property
    def is_variable(self):
        return self.uid.startswith("${") and self.uid.endswith("}")

    def update(self, ds_map: Mapping[str, DSRef], strict=False) -> bool:
        if self.is_variable:
            return False

        if self.uid not in ds_map:
            if strict:
                raise UnmappedDatasourceError(self.uid)
            return False

        self.uid = ds_map[self.uid].uid
        return True


class Target(BaseModel):
    datasource: DataSource | None = None

    @property
    def all_datasources(self) -> Generator[DataSource, None, None]:
        if self.datasource is not None:
            yield self.datasource


class Panel(BaseModel):
    datasource: DataSource | str | None = None
    targets: list[Target] | None = None
    panels: list["Panel"] | None = None

    model_config = ConfigDict(extra="allow")

    def upgrade_datasource(self, ds_config: Mapping[str, DataSource]) -> None:
        """Upgrade string-datasource into an datasource object.

        Older Grafana versions used to put a string (name) into the datasource field.
        """
        if isinstance(self.datasource, str):
            self.datasource = ds_config[self.datasource]

    @property
    def all_datasources(self) -> Generator[DataSource, None, None]:
        if self.datasource is not None:
            if isinstance(self.datasource, str):
                msg = f"please run upgrade_datasource to resolve datasource `{self.datasource}`"
                raise ValueError(msg)
            else:
                yield self.datasource

        if self.targets is not None:
            for t in self.targets:
                yield from t.all_datasources

        if self.panels is not None:
            for p in self.panels:
                yield from p.all_datasources

    def update_datasources(
        self,
        ds_map: Mapping[str, DSRef],
        strict=False,
    ) -> int:
        ct = 0

        for ds in self.all_datasources:
            if ds.update(ds_map, strict):
                ct += 1

        return ct


class TemplatingItemCurrent(BaseModel):
    text: str | list[str] | None = Field(default=None)
    value: str | list[str] | None = Field(default=None)

    def update_datasource(self, ds_map: Mapping[str, DSRef], strict=False) -> bool:
        if not isinstance(self.text, str):
            return False

        if not isinstance(self.value, str):
            return False

        if self.value not in ds_map:
            if strict:
                raise UnmappedDatasourceError(self.value)
            return False

        self.text = ds_map[self.value].name
        self.value = ds_map[self.value].uid

        return True


class TemplatingItem(BaseModel):
    current: TemplatingItemCurrent | None = None
    datasource: DataSource | None = None
    type_: str = Field(alias="type")

    model_config = ConfigDict(extra="allow")

    @property
    def all_datasources(self) -> Generator[DataSource, None, None]:
        if self.datasource is not None:
            yield self.datasource

    @property
    def is_datasource(self) -> bool:
        return self.type_ == "datasource"


class Templating(BaseModel):
    list_: list[TemplatingItem] | None = Field(alias="list", default=None)

    model_config = ConfigDict(extra="allow")

    @property
    def all_datasources(self) -> Generator[DataSource, None, None]:
        if self.list_ is not None:
            for t in self.list_:
                yield from t.all_datasources

    def update_datasources(self, ds_map: Mapping[str, DSRef], strict=False) -> int:
        if self.list_ is None:
            return 0

        ct = 0

        for t in self.list_:
            if not t.is_datasource or t.current is None:
                continue

            if t.current.update_datasource(ds_map, strict):
                ct += 1

        return ct


class DashboardData(BaseModel):
    uid: str
    title: str
    version: int | None = None

    panels: list[Panel] | None = None
    templating: Templating | None = None

    model_config = ConfigDict(extra="allow")

    @property
    def all_datasources(self) -> Generator[DataSource, None, None]:
        if self.panels is not None:
            for p in self.panels:
                yield from p.all_datasources

        if self.templating is not None:
            yield from self.templating.all_datasources

    @property
    def datasource_count(self) -> int:
        return len(list(self.all_datasources))

    @property
    def variable_datasource_count(self) -> int:
        return len([ds for ds in self.all_datasources if ds.is_variable])

    def upgrade_datasources(self, ds_config: Mapping[str, DataSource]) -> None:
        if self.panels is not None:
            for p in self.panels:
                p.upgrade_datasource(ds_config)

    def update_datasources(
        self,
        ds_map: Mapping[str, DSRef],
        strict=False,
    ) -> int:
        ct = 0

        if self.panels is not None:
            for p in self.panels:
                ct += p.update_datasources(ds_map, strict)

        if self.templating is not None:
            ct += self.templating.update_datasources(ds_map, strict)

        return ct



import importlib.resources

import pytest

from grafana_sync.api.models import GetDashboardResponse

from . import responses


def read_response(filename: str) -> GetDashboardResponse:
    ref = importlib.resources.files(responses) / filename
    with importlib.resources.as_file(ref) as path, open(path, "rb") as f:
        return GetDashboardResponse.model_validate_json(f.read())


@pytest.mark.parametrize(
    ("filename"),
    [
        ("get-dashboard-datasource-string.json"),
    ],
)
def test_read(filename):
    read_response(filename)



import tempfile
from pathlib import Path

import pytest

from grafana_sync.api.client import GrafanaClient
from grafana_sync.api.models import DashboardData
from grafana_sync.backup import GrafanaBackup
from grafana_sync.restore import GrafanaRestore

pytestmark = pytest.mark.docker


@pytest.fixture
def backup_dir():
    """Create a temporary directory for backup/restore testing."""
    with tempfile.TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)


async def test_restore_folder(grafana: GrafanaClient, backup_dir: Path):
    backup = GrafanaBackup(grafana, backup_dir)
    restore = GrafanaRestore(grafana, backup_dir)

    # Create and backup a test folder
    folder_uid = "test-restore-folder"
    folder_title = "Test Restore Folder"
    await grafana.create_folder(title=folder_title, uid=folder_uid)
    await backup.backup_folder(folder_uid)

    # Delete the original folder
    await grafana.delete_folder(folder_uid)

    # Restore the folder
    await restore.restore_folder(folder_uid)

    # Verify restored folder
    folder = await grafana.get_folder(folder_uid)
    assert folder.uid == folder_uid
    assert folder.title == folder_title


async def test_restore_dashboard(grafana: GrafanaClient, backup_dir: Path):
    backup = GrafanaBackup(grafana, backup_dir)
    restore = GrafanaRestore(grafana, backup_dir)

    # Create and backup a test dashboard
    dashboard = DashboardData(
        uid="test-restore-dashboard", title="Test Restore Dashboard"
    )

    await grafana.update_dashboard(dashboard)
    await backup.backup_dashboard("test-restore-dashboard")

    # Delete the original dashboard
    await grafana.delete_dashboard("test-restore-dashboard")

    # Restore the dashboard
    await restore.restore_dashboard("test-restore-dashboard")

    # Verify restored dashboard
    restored = await grafana.get_dashboard("test-restore-dashboard")
    assert restored.dashboard.uid == "test-restore-dashboard"
    assert restored.dashboard.title == "Test Restore Dashboard"


async def test_restore_recursive(grafana: GrafanaClient, backup_dir: Path):
    backup = GrafanaBackup(grafana, backup_dir)
    restore = GrafanaRestore(grafana, backup_dir)

    # Create test structure
    folder_uid = "test-restore-recursive"
    await grafana.create_folder(title="Test Restore Recursive", uid=folder_uid)

    dashboard = DashboardData(
        uid="test-restore-dash-recursive", title="Test Restore Dashboard Recursive"
    )
    await grafana.update_dashboard(dashboard, folder_uid=folder_uid)

    # Backup everything
    await backup.backup_recursive()

    # Delete everything
    await grafana.delete_dashboard("test-restore-dash-recursive")
    await grafana.delete_folder(folder_uid)

    # Restore everything
    await restore.restore_recursive()

    # Verify folder was restored
    folder = await grafana.get_folder(folder_uid)
    assert folder.uid == folder_uid
    assert folder.title == "Test Restore Recursive"

    # Verify dashboard was restored
    dashboard = await grafana.get_dashboard("test-restore-dash-recursive")
    assert dashboard.dashboard.uid == "test-restore-dash-recursive"
    assert dashboard.dashboard.title == "Test Restore Dashboard Recursive"


async def test_restore_recursive_top_level_dashboard(
    grafana: GrafanaClient, backup_dir: Path
):
    backup = GrafanaBackup(grafana, backup_dir)
    restore = GrafanaRestore(grafana, backup_dir)

    # Create test structure
    dashboard = DashboardData(
        uid="test-restore-dash-recursive", title="Test Restore Dashboard Recursive"
    )
    await grafana.update_dashboard(dashboard)

    # Backup everything
    await backup.backup_recursive()

    # Delete everything
    await grafana.delete_dashboard("test-restore-dash-recursive")

    # Restore everything
    await restore.restore_recursive()

    # Verify dashboard was restored
    dashboard = await grafana.get_dashboard("test-restore-dash-recursive")
    assert dashboard.dashboard.uid == "test-restore-dash-recursive"
    assert dashboard.dashboard.title == "Test Restore Dashboard Recursive"
